package com.wendall.controllers 
{
	import com.wendall.events.WendallEvent;
	import com.wendall.events.WendallNavEvent;
	import com.wendall.views.MainView;
	
	import flash.display.MovieClip;
	import flash.display.StageAlign;
	import flash.display.StageScaleMode;
	import flash.events.Event;	

	/**
	 * @author Brian Kuzma
	 */
	public class AppController extends MovieClip
	{
		private var mainView:MainView = new MainView();
		private var bioController:BioController = new BioController();
		private var homeController:HomeController = new HomeController();
		private var newsController:NewsController = new NewsController();
		private var sectionControllers:Array = [bioController, homeController, newsController];
		private var activeSectionController:SectionController;

		public function AppController()
		{
			trace("AppController::AppController");
			addEventListener(Event.ADDED_TO_STAGE, onAddedToStage);
		}
		
		private function onAddedToStage(event:Event):void
		{
			init();
		}
		
		public function init():void
		{
			trace("AppController::init");
			
			stage.scaleMode = StageScaleMode.NO_SCALE;
			stage.align = StageAlign.TOP_LEFT;
			
			initViews();
			initListeners();
			
			mainView.addEventListener(WendallEvent.COMPLETE_INTRO, onCompleteIntro);
			mainView.intro();
		}
		
		private function initViews():void
		{
			addChild(mainView);
			for (var i:uint = 0; i < sectionControllers.length; i++)
			{
				var sectionController:SectionController = sectionControllers[i];
				mainView.addView(sectionController.view);
			}
		}

		private function initListeners():void
		{
			mainView.addEventListener(WendallNavEvent.CLICK_NAV, onClickNav);
		}

		private function activateSection(sectionController:SectionController):void
		{
			trace("AppController::activateSection -->" + sectionController);
			sectionController.addEventListener(WendallEvent.ACTIVATED, onSectionActivated);
			sectionController.activateSection();
		}

		private function onCompleteIntro(event:WendallEvent):void
		{
			mainView.removeEventListener(WendallEvent.COMPLETE_INTRO, onCompleteIntro);
			mainView.activateNav();
			
			activateSection(homeController);
		}
		
		private function onClickNav(event:WendallNavEvent):void
		{
			// deactivate nav
			mainView.deactivateNav();
			
			var sectionToClose:SectionController = activeSectionController;

			// set new section
			switch (event.sectionClicked)
			{
				case WendallNavEvent.HOME:
					activeSectionController = homeController;
					break;
				case WendallNavEvent.BIO:
					activeSectionController = bioController;
					break;
				case WendallNavEvent.NEWS:
					activeSectionController = newsController;
					break;
			}
			
			// close current section
			if (sectionToClose)
			{
				sectionToClose.addEventListener(WendallEvent.DEACTIVATED, onSectionDeactivated);
				sectionToClose.deactivateSection();
			}
			else
			{
				activateSection(activeSectionController);
			}
			
		}
		
		private function onSectionDeactivated(event:WendallEvent):void
		{
			// activate section
			activateSection(activeSectionController);
		}
		
		private function onSectionActivated(event:WendallEvent):void
		{
			// reactivate nav
			mainView.activateNav();
		}
	}
}